@@ -23,20 +23,20 @@
 #include <XSUB.h>
 
 
+#if 0
 #ifdef PERL_IMPLICIT_CONTEXT
 #  error "PERL_IMPLICIT_CONTEXT is not fully supported."
 #endif
+#endif
 
 #define PICKLE_INTERPRETER_PRIVATE				\
     Interpreter (Interpreter_imp* i) : interpreter_imp (i) {}	\
-    static Interpreter& ref (Interpreter_imp*& ptr)		\
-    { return *(Interpreter*) &ptr; }				\
 								\
     void init (int, const char* const *, const char* const *);	\
     Scalar_imp* call_function (Scalar_imp* func, int argc,	\
 			       Scalar_imp** argv,		\
 			       Context ctx) const;		\
-    static void propagate_to_perl (Exception* e);		\
+    static void propagate_to_perl (pTHX_ Exception* e);		\
     static XS (xs_entry_one_arg);				\
     static XS (xs_entry_hashref);				\
     static XS (xs_entry_list);
@@ -53,6 +53,7 @@ void my_sv_refcnt_dec (SV* sv) { my_inline_sv_refcnt_dec (sv); }
 namespace Pickle
 {
 
+  // XXX need callbacks to take an interp.
   static Scalar
   my_warner (Scalar& msg)
   {
@@ -75,7 +76,7 @@ namespace Pickle
     // XXX Perl doesn't assign early enough?!
     if (PL_warnhook)
       SvREFCNT_dec (PL_warnhook);
-    PL_warnhook = Perl_newSVpv ("Pickle::my_warner", 0);
+    PL_warnhook = newSVpv ("Pickle::my_warner", 0);
   }
 
   static void barf (PerlInterpreter* p, const char* whodied, int status)
@@ -105,14 +106,6 @@ namespace Pickle
     if (status)
       barf (my_perl, "perl_parse", status);
 
-    // Avoid perl_call_method because it cannot trap the no-such-method
-    // error.  XXX
-    eval_string
-      ("sub Pickle::call_method {		\
-            my ($meth, $obj, $args) = @_;	\
-            $obj->$meth (@$args);		\
-        }");
-
     status = perl_run (my_perl);
     if (status)
       barf (my_perl, "perl_run", status);
@@ -129,90 +122,182 @@ namespace Pickle
     static const char* argv[] = { "Pickle", "-e0" };
     init (sizeof argv / sizeof argv [0], argv, 0);
   }
-  Interpreter::Interpreter (const vector<string>& args)
+  Interpreter::Interpreter (const vector<const string>& args)
   {
     int argc = args.size ();
     const char* argv[argc];
-    vector<string>::const_iterator i = args.begin ();
+    vector<const string>::const_iterator i = args.begin ();
     
     for (int n = 0; n < argc; n++)
       argv [n] = (i++)->c_str();
     init (argc, argv, 0);
   }
-  Interpreter::Interpreter (const vector<string>& args,
-			    const vector<string>& env)
+  Interpreter::Interpreter (const vector<const string>& args,
+			    const vector<const string>& env)
   {
     int argc = args.size ();
     int envc = env.size ();
-    const char* argv[argc];
-    vector<string>::const_iterator i = args.begin();
-    
-    for (int n = 0; n < argc; n++)
-      argv [n] = (i++)->c_str();
-    init (argc, argv, 0);
+    const char* argv[argc + 1];  // XXX portable?
+    const char* envp[envc + 1];  // XXX portable?
+    vector<const string>::const_iterator p;
+    int i;
+
+    for (p = args .begin (), i = 0; i < argc; i++)
+      argv [i] = (p++) ->c_str ();
+    argv [argc] = 0;
+    for (p = env .begin (), i = 0; i < envc; i++)
+      envp [i] = (p++) ->c_str ();
+    envp [envc] = 0;
+    init (argc, argv, envp);
   }
 
   Interpreter::~Interpreter ()
   {
     perl_destruct (my_perl);
     perl_free (my_perl);
-    PL_curinterp = 0;
+    PERL_SET_CONTEXT (0);
   }
 
-  const Interpreter&
+  Interpreter*
   Interpreter::get_current ()
   {
-    return Interpreter::ref (PL_curinterp);
+    // XXX use PERL_GET_CONTEXT and keep a table of context --> Interpreter.
+#ifdef PERL_IMPLICIT_CONTEXT
+    dTHX;
+    return aTHX ?
+      // XXX leak memory for now, just testing this stuff...
+      new Interpreter (aTHX)
+      : 0;
+#else  // !PERL_IMPLICIT_CONTEXT
+    return PL_curinterp ? (Interpreter*) &PL_curinterp : 0;
+#endif
   }
 
-  bool
-  Interpreter::ping ()
+  // Data allocation.  cf. "Conversion from native C++ types to scalar"
+  // in scalar.cc.
+
+  Pickle::Scalar Interpreter::Scalar () const
+  { return &PL_sv_undef; }
+  Pickle::Scalar Interpreter::Scalar (unsigned long i) const
+  { return newSVnv (i); }
+  Pickle::Scalar Interpreter::Scalar (long i) const
+  { return newSViv (i); }
+  Pickle::Scalar Interpreter::Scalar (unsigned int i) const
+  { return newSVnv (i); }
+  Pickle::Scalar Interpreter::Scalar (int i) const
+  { return newSViv (i); }
+  Pickle::Scalar Interpreter::Scalar (unsigned short i) const
+  { return newSVnv (i); }
+  Pickle::Scalar Interpreter::Scalar (short i) const
+  { return newSViv (i); }
+  Pickle::Scalar Interpreter::Scalar (unsigned char i) const
+  { return newSVnv (i); }
+  Pickle::Scalar Interpreter::Scalar (signed char i) const
+  { return newSViv (i); }
+
+  Pickle::Scalar
+  Interpreter::Scalar (const string& s) const
   {
-    dTHX;
-    return PL_curinterp != 0;
+    return newSVpvn (s .data (), s .size ());
   }
 
+  Pickle::Scalar
+  Interpreter::Scalar (const char* s) const
+  {
+    return newSVpv (s, 0);
+  }
+
+  Pickle::Scalar
+  Interpreter::Scalar (const char* s, unsigned long len) const
+  {
+    return newSVpvn (s, len);
+  }
+
+  Pickle::Scalar Interpreter::Scalar (double d) const
+  { return newSVnv (d); }
+  Pickle::Scalar Interpreter::Scalar (float d) const
+  { return newSVnv (d); }
+
+  Pickle::Scalar
+  Interpreter::Scalar (bool b) const
+  {
+    return b ? &PL_sv_yes : &PL_sv_no;
+  }
+
+  Pickle::Scalarref Interpreter::Scalarref () const
+  { return Pickle::Scalarref (newRV_noinc (newSVsv (&PL_sv_undef)), false);}
+  Pickle::Arrayref Interpreter::Arrayref () const
+  { return Pickle::Arrayref (newRV_noinc ((SV*) newAV ()), false); }
+  Pickle::Hashref Interpreter::Hashref () const
+  { return Pickle::Hashref (newRV_noinc ((SV*) newHV ()), false); }
+  Pickle::List Interpreter::List () const
+  { return Arrayref (); }
+
+  // Symbol table lookups.
+
+  Pickle::Scalarref Interpreter::Scalarref (const string& name)
+  { return Scalarref (name .c_str ()); }  // XXX embedded \0
+  Pickle::Scalarref Interpreter::Scalarref (const char* name)
+  { return Pickle::Scalarref (newRV (get_sv (name, 1)), false); }
+  Pickle::Arrayref Interpreter::Arrayref (const string& name)
+  { return Arrayref (name .c_str ()); }  // XXX embedded \0
+  Pickle::Arrayref Interpreter::Arrayref (const char* name)
+  { return Pickle::Arrayref (newRV ((SV*) get_av (name, 1)), false); }
+  Pickle::Hashref Interpreter::Hashref (const string& name)
+  { return Hashref (name .c_str ()); }  // XXX embedded \0
+  Pickle::Hashref Interpreter::Hashref (const char* name)
+  { return Pickle::Hashref (newRV ((SV*) get_hv (name, 1)), false); }
+  Pickle::Coderef Interpreter::Coderef (const string& name)
+  { return Coderef (name .c_str ()); }  // XXX embedded \0
+  Pickle::Coderef Interpreter::Coderef (const char* name)
+  { return Pickle::Coderef (newRV ((SV*) get_cv (name, 1)), false); }
+
+  // Compilation.
+
   void
   Interpreter::require_module (const string& bare) const
   {
     eval_string (string ("require ") .append (bare));
   }
 
-  Scalar
+  Pickle::Scalar
   Interpreter::eval_string (const string& proggie) const
   {
     djSP;
     SV* retsv;
-    Scalar err;
+    Pickle::Scalar err;
 
     ENTER;
     SAVETMPS;
-    // This is equivalent to `local($@)'.
-    Perl_save_scalar (PL_errgv);
+    // This is equivalent to `local($@)'.  XXX missing from perlapi.pod
+    save_scalar (PL_errgv);
 
-    retsv = Perl_eval_pv (proggie.c_str(), G_SCALAR);
+    retsv = eval_pv (proggie.c_str(), G_SCALAR);
 
     // I wasted half a day figuring out that this is needed.
     POPMARK;
 
     SvREFCNT_inc (retsv);
     if (SvTRUE (ERRSV))
-      err = Perl_newSVsv (ERRSV);
+      err = newSVsv (ERRSV);
 
     FREETMPS;
     LEAVE;
 
     if (err)
-      die (err);
+      throw new Exception (err);
 
     return retsv;
   }
 
-  Scalar
-  Interpreter::call_function (const Scalar& func, const List& args,
+  // Transfering control from C++ to Perl.
+
+  Pickle::Scalar
+  Interpreter::call_function (const Pickle::Scalar& func,
+			      const Pickle::List& args,
 			      Context cx = SCALAR) const
   {
-    AV* av = (AV*) SvRV (((const Arrayref&) args) .imp);
+    AV* av = (AV*) SvRV (((const Pickle::Arrayref&) args) .imp);
     return call_function (const_cast<SV*> (func .imp), 1 + AvFILL (av),
 			  AvARRAY (av), cx);
   }
@@ -222,7 +307,7 @@ namespace Pickle
   {
     djSP;
     SV* retsv;
-    Scalar err;
+    Pickle::Scalar err;
     I32 numret;
     I32 ctx;
 
@@ -236,8 +321,8 @@ namespace Pickle
 
     ENTER;
     SAVETMPS;
-    // This is equivalent to `local($@)'.
-    Perl_save_scalar (PL_errgv);
+    // This is equivalent to `local($@)'.  XXX missing from perlapi.pod
+    save_scalar (PL_errgv);
 
     PUSHMARK (sp);
     EXTEND (sp, argc);
@@ -245,14 +330,14 @@ namespace Pickle
       PUSHs (argv [i]);
     PUTBACK;
 
-    numret = perl_call_sv (func, ctx | G_EVAL);
+    numret = call_sv (func, ctx | G_EVAL);
     SPAGAIN;
 
     switch (ctx)
       {
       case G_ARRAY:
 	sp -= numret;
-	retsv = Perl_newRV_noinc ((SV*) Perl_av_make (numret, sp + 1));
+	retsv = newRV_noinc ((SV*) av_make (numret, sp + 1));
 	break;
 
       case G_SCALAR:
@@ -267,35 +352,38 @@ namespace Pickle
     PUTBACK;
 
     if (SvTRUE (ERRSV))
-      err = Perl_newSVsv (ERRSV);
+      err = newSVsv (ERRSV);
 
     FREETMPS;
     LEAVE;
 
     if (err)
-      die (err);
+      throw new Exception (err);
 
     return retsv;
   }
 
-  void Interpreter::propagate_to_perl (Exception* e)
+  // Tranfering control from Perl to C++.
+
+  void
+  Interpreter::propagate_to_perl (pTHX_ Exception* e)
   {
     // XXX This is not especially clever.
-    Perl_sv_setsv (ERRSV, e->get_scalar () .imp);
+    sv_setsv (ERRSV, e->get_scalar () .imp);
     delete e;
-    Perl_croak ("%_", ERRSV);
+    croak ("%_", ERRSV);
   }
 
   XS (Interpreter::xs_entry_one_arg)
   {
     dXSARGS;
     if (items != 1)
-      Perl_croak ("Usage: %s(arg)", GvNAME (CvGV (cv)));
+      croak ("Usage: %s(arg)", GvNAME (CvGV (cv)));
 
     try
       {
-	Scalar arg (SvREFCNT_inc (ST (0)));
-	Scalar ret (((sub_one_arg) CvXSUBANY (cv) .any_ptr) (arg));
+	Pickle::Scalar arg (SvREFCNT_inc (ST (0)));
+	Pickle::Scalar ret (((sub_one_arg) CvXSUBANY (cv) .any_ptr) (arg));
 	/* The callback return value is basically a time bomb, an SV with
 	   0 refcount: actually, its refcnt is about to drop to 0 when the
 	   container (Scalar) dies.  Therefore, we must inc its refcount.
@@ -304,11 +392,11 @@ namespace Pickle
 	   with another.  A clever compiler might optimize away SvREFCNT_inc.
 	   (nah...)
 	*/
-	ST (0) = SvREFCNT_inc (Perl_sv_2mortal (ret .imp));
+	ST (0) = SvREFCNT_inc (sv_2mortal (ret .imp));
       }
     catch (Exception* e)
       {
-	propagate_to_perl (e);
+	propagate_to_perl (aTHX_ e);
       }
     XSRETURN (1);
   }
@@ -317,22 +405,23 @@ namespace Pickle
   {
     dXSARGS;
     if ((items % 2) == 0)
-      Perl_croak ("Usage: OBJECT->%s (NAME, VALUE, ...)", GvNAME (CvGV (cv)));
+      croak ("Usage: OBJECT->%s (NAME, VALUE, ...)", GvNAME (CvGV (cv)));
 
-    HV* args = Perl_newHV ();
+    HV* args = newHV ();
     for (int i = 1; i < items - 1; i += 2)
-      Perl_hv_store_ent (args, ST (i), SvREFCNT_inc (ST (i + 1)), 0);
+      hv_store_ent (args, ST (i), SvREFCNT_inc (ST (i + 1)), 0);
 
     try
       {
-	Scalar obj (SvREFCNT_inc (ST (0)));
-	Hashref arg (Perl_newRV_noinc ((SV*) args));
-	Scalar ret (((sub_hashref) CvXSUBANY (cv) .any_ptr) (obj, arg));
-	ST (0) = SvREFCNT_inc (Perl_sv_2mortal (ret .imp));
+	Pickle::Scalar obj (SvREFCNT_inc (ST (0)));
+	Pickle::Hashref arg (newRV_noinc ((SV*) args), false);
+	Pickle::Scalar ret (((sub_hashref) CvXSUBANY (cv) .any_ptr)
+			    (obj, arg));
+	ST (0) = SvREFCNT_inc (sv_2mortal (ret .imp));
       }
     catch (Exception* e)
       {
-	propagate_to_perl (e);
+	propagate_to_perl (aTHX_ e);
       }
     XSRETURN (1);
   }
@@ -344,7 +433,7 @@ namespace Pickle
     SV* retsv;
     AV* retav;
 
-    SV* args = Perl_newRV_noinc ((SV*) Perl_av_make (items, & ST (0)));
+    SV* args = newRV_noinc ((SV*) av_make (items, & ST (0)));
     switch (GIMME_V)
       {
       case G_ARRAY : cx = LIST;   break;
@@ -354,13 +443,13 @@ namespace Pickle
 
     try
       {
-	List arglist (args);
-	List retlist (((sub) CvXSUBANY (cv) .any_ptr) (arglist, cx));
-	retsv = SvREFCNT_inc (Perl_sv_2mortal (((Arrayref&) retlist) .imp));
+	Pickle::List arglist (args);
+	Pickle::List retlist (((sub) CvXSUBANY (cv) .any_ptr) (arglist, cx));
+	retsv = SvREFCNT_inc (sv_2mortal (((Pickle::Arrayref&) retlist) .imp));
       }
     catch (Exception* e)
       {
-	propagate_to_perl (e);
+	propagate_to_perl (aTHX_ e);
       }
 
     retav = (AV*) SvRV (retsv);
@@ -369,7 +458,7 @@ namespace Pickle
       {
       case LIST:
 	{
-	  size_t sz = Perl_av_len (retav) + 1;
+	  size_t sz = av_len (retav) + 1;
 	  if (sz != 0)
 	    {
 	      if (sz > (size_t) items)
@@ -378,7 +467,7 @@ namespace Pickle
 	      // Note: List AVs are never magical.
 	      SV** svp = AvARRAY (retav);
 	      for (size_t i = 0; i < sz; i++)
-		ST (i) = SvREFCNT_inc (Perl_sv_2mortal (svp [i]));
+		ST (i) = SvREFCNT_inc (sv_2mortal (svp [i]));
 	    }
 	  XSRETURN (sz);
 	}
@@ -386,13 +475,13 @@ namespace Pickle
 
       case SCALAR:
 	{
-	  size_t sz = Perl_av_len (retav) + 1;
+	  size_t sz = av_len (retav) + 1;
 	  if (sz != 0)
 	    {
 	      if (items < 1)
 		EXTEND (sp, 1);
 
-	      ST (0) = SvREFCNT_inc (Perl_sv_2mortal
+	      ST (0) = SvREFCNT_inc (sv_2mortal
 				     (AvARRAY (retav) [sz - 1]));
 	      XSRETURN (1);
 	    }
@@ -407,12 +496,12 @@ namespace Pickle
   }
 
   static void
-  reg (const string& package, const string& name, void* fn, XS (xs))
+  reg (pTHX_ const string& package, const string& name, void* fn, XS (xs))
   {
     string fullname (package);
     fullname .append ("::") .append (name);
-    CV* cv = Perl_newXS (const_cast<char*> (fullname .c_str ()),
-			 xs, const_cast<char*> (__FILE__));
+    CV* cv = newXS (const_cast<char*> (fullname .c_str ()),
+		    xs, const_cast<char*> (__FILE__));
     CvXSUBANY (cv) .any_ptr = fn;
   }
 
@@ -420,33 +509,35 @@ namespace Pickle
   Interpreter::define_sub (const string& package, const string& name,
 			   sub_one_arg fn) const
   {
-    reg (package, name, fn, xs_entry_one_arg);
+    reg (aTHX_ package, name, fn, xs_entry_one_arg);
   }
 
   void
   Interpreter::define_sub (const string& package, const string& name,
 			   sub fn) const
   {
-    reg (package, name, fn, xs_entry_list);
+    reg (aTHX_ package, name, fn, xs_entry_list);
   }
 
   void
   Interpreter::define_sub (const string& package, const string& name,
 			   sub_hashref fn) const
   {
-    reg (package, name, fn, xs_entry_hashref);
+    reg (aTHX_ package, name, fn, xs_entry_hashref);
   }
 
+  // Exception class.
+
   Exception::Exception (const string& s)
   {
     Interpreter::vivify ();
-    scalar = s;
+    msg = s;
   }
 
   Exception::Exception (const char* s)
   {
     Interpreter::vivify ();
-    scalar = s;
+    msg = s;
   }
 
 }
