@@ -28,7 +28,7 @@ namespace Pickle
   lookup_scalar (const string& name)
   {
     dTHX;
-    return Perl_newRV ((SV*) Perl_get_sv (name .c_str (), 1));
+    return newRV ((SV*) get_sv (name .c_str (), 1));
   }
 
   Scalarref::Scalarref (const string& name) : Scalar (lookup_scalar (name)) {}
@@ -37,15 +37,15 @@ namespace Pickle
   Scalar
   Scalarref::fetch () const
   {
-    dTHX;
+    dInterp;
     return SvRV (imp);
   }
 
   void
   Scalarref::store (const Scalar& v)
   {
-    dTHX;
-    Perl_sv_setsv (SvRV (imp), v.imp);
+    dInterp;
+    sv_setsv (SvRV (imp), v.imp);
   }
 
 }
